"""
	Mesh Module

"""
import Blender
import math3d

MDEBUG = 0

def getMedianPoint( l ):
	def addv(a,b):return a+b
	return (reduce( addv, [math3d.Vector(v.co) for v in l] )) / float(len(l))

class LinkedEdge:
	def __init__( self, i, me, a, b ):
		self.mesh = me
		self.pa = a
		self.pb = b
		self.mp = None
		self.faces = []
		self.id = i
	
	def calcuMidPoint( self ):
		if not self.mp:
			self.mp = self.mesh.addVertex( math3d.Vector( self.pa.co ).Lerp( math3d.Vector( self.pb.co ), 0.5 ) )
		return self.mp
	
	def getOtherFace( self, face ):
		if self.faces.count( face ) and len(self.faces) < 2:return None
		return self.faces[ not self.faces.index(face) ]

class FaceWrapper:
	all_id = 0
	
	def __init__( self, me, bf ):
		global MDEBUG
		self.mesh = me
		self.bface = bf
		self.edges = []
		self.id = FaceWrapper.all_id
		FaceWrapper.all_id += 1
		if MDEBUG:
			print "Construct Face(",self,")"
			print "\tFace:[",bf,"]"
			print "\tID:[%d]" % self.id
	
	def isSelected( self ):
		global MDEBUG
		if MDEBUG:
			print "\tFlag %d" % self.bface.flag
		return self.bface.flag & 1
	
	def getMedianPoint( self ):
		def addv(a,b):return a+b
		return (reduce( addv, [math3d.Vector(v.co) for v in self.bface.v] )) / float(len(self.bface.v))
	
	def registerEdge( self, e ):
		global MDEBUG
		if not self.edges.count( e ):
			self.edges.append( e )
			if MDEBUG:
				print "\tRegister Edge:[EdgeID %d]->[FaceID %d]" %(e.id,self.id)
		if not e.faces.count( self ):
			e.faces.append( self )
			if MDEBUG:
				print "\tRegister Face:[FaceID %d]->[EdgeID %d]" %(self.id,e.id)
	
	def __cmp__(self, o ):
		if len(self.bface.v) != len(o.bface.v):
			return -1
		for v in self.bface.v:
			if not v in o.bface.v:return -1
		return 0
	
	def getNormal(self):
		V = math3d.Vector
		pa = V(self.bface.v[0].co)
		pb = V(self.bface.v[1].co)
		pc = V(self.bface.v[2].co)
		c1 = (pa-pb).Cross( pc-pb )
		if len(self.bface.v) == 4:
			pd = V(self.bface.v[3].co)
			c2 = (pc-pd).Cross(pa-pd)
			c1 = (c1 + c2) * 0.5
			return c1.Normalize()
		return c1.Normalize()
	
	def getSharedEdgeID( self, other ):
		for edge in other.edges:
			if self.edges.count( edge ):
				return self.edges.index( edge )
	
	def getAdjacentFace( self, id ):
		return self.edges[id].getOtherFace(self)

class MeshWrapper:
	def __init__( self, me ):
		self.nmesh = me
		self.edges = {}
		self.faces = []
		self.numv = len(me.verts)
		self.init()
	
	def getFaceFromNMFace( self, f ):
		e1 = self.getEdge( f.v[0], f.v[1] )
		e2 = self.getEdge( f.v[1], f.v[2] )
		return e1.faces[ e1.faces[0] != e2.faces[0] ]
	
	def getSelectedFaces(self):
		return [f for f in self.faces if f.isSelected()]
	
	def getEdge( self, bv1, bv2 ):
		global MDEBUG
		if MDEBUG:print "\tGet Edge Method"
		id = bv1.index << 16 | bv2.index
		if self.edges.has_key(id):
			if MDEBUG:
				print "\t\t.1 DictID %d" % id
			return self.edges[id]
		id = bv2.index << 16 | bv1.index
		if self.edges.has_key(id):
			if MDEBUG:
				print "\t\t.2 DictID %d" % id
			return self.edges[id]
		return self.createEdge(bv1,bv2)
	
	def createEdge(self,bv1,bv2):
		global MDEBUG
		if MDEBUG:print "\t\tCreate Ned Edge"
		i = bv1.index << 16 | bv2.index
		ret = LinkedEdge(i, self, bv1,bv2)
		self.edges[ i ] = ret
		return ret
	
	def init(self):
		if self.faces or self.edges:
			self.faces = []
			self.edges = {}
		for face in self.nmesh.faces:
			self.registerFace( face )

	def addVertex( self, co ):
		global MDEBUG
		if MDEBUG:
			print "\t\tCreate Ned Vertex %s" % co
		v = Blender.NMesh.Vert( co.x, co.y, co.z )
		v.index = len(self.nmesh.verts)
		self.nmesh.verts.append(v)
		return v
	
	def addFaceFromVertices( self, v ):
		global MDEBUG
		if MDEBUG:
			print "\t\tCreate New Face"
		f = Blender.NMesh.Face()
		for vv in v:
			f.v.append( vv )
		self.nmesh.faces.append( f )
		return self.registerFace( f )
	
	def addFaceFromEdges( self, e ):
		self.addFaceFromVertices( getVertexStreamFromEdges( e ) )
		
	def registerFace( self, f ):
		fw = FaceWrapper(self,f)
		for i in range(len(f.v)):
			e = self.getEdge( f.v[i], f.v[(i+1)%len(f.v)] )
			fw.registerEdge( e )
		self.faces.append(fw)
		return fw
	
	def deleteFace( self, f ):
		for e in f.edges:
			e.faces.remove( f )
		self.deleteOnlyFace(f)
	
	def deleteOnlyFace( self, f ):
		if self.nmesh.faces.count( f.bface ):
			self.nmesh.faces.remove( f.bface )
		self.faces.remove( f )
	
	def toBlender( self ):
		n = self.nmesh.name
		self.nmesh.update()
		Blender.Redraw()
	
	def setFaceFlag( self, face, val ):
		face.bface.flag = val
	
	def update( self ):
		me = Blender.NMesh.GetRaw( self.nmesh.name )
		if self.numv != len(me.verts) or \
			len(self.faces) != len(me.faces):
			self.init()
			return
		for i in range(len(me.faces)):
			self.faces[i].bface.flag = me.faces[i].flag